// A comparable, immutable data type that represents a country by its name, capital, and population.

import stdlib.StdOut;

import java.util.Comparator;

public class Country implements Comparable<Country> {
    private String name; // name
    private String capital; // capital
    private long population; // population

    // Constructs a country given its name, capital, and population.
    public Country(String name, String capital, long population) {
        this.name = name;
        this.capital = capital;
        this.population = population;
    }

    // Returns a string representation of this country.
    public String toString() {
        return this.name + " (" + this.capital + "): " + this.population;
    }

    // Returns a comparison (-ve, 0, +ve integer) of this and the other country, by name.
    public int compareTo(Country other) {
        return this.name.compareTo(other.name);
    }

    // Returns a comparator for comparing two countries by capital.
    public static Comparator<Country> capitalOrder() {
        return new CapitalOrder();
    }

    // Returns a comparator for comparing two countries by population.
    public static Comparator<Country> populationOrder() {
        return new PopulationOrder();
    }

    // Returns a reverse comparator for comparing two countries by population.
    public static Comparator<Country> reversePopulationOrder() {
        return new ReversePopulationOrder();
    }

    // A comparator for comparing two countries by capital.
    private static class CapitalOrder implements Comparator<Country> {
        // Returns a comparison (-ve, 0, +ve integer) of countries a and b, by capital.
        public int compare(Country a, Country b) {
            return a.capital.compareTo(b.capital);
        }
    }

    // A comparator for comparing two countries by population.
    private static class PopulationOrder implements Comparator<Country> {
        // Returns a comparison (-ve, 0, +ve integer) of countries a and b, by population.
        public int compare(Country a, Country b) {
            return Long.compare(a.population, b.population);
        }
    }

    // A reverse comparator for comparing two countries by population.
    private static class ReversePopulationOrder implements Comparator<Country> {
        // Returns a reverse comparison (+ve, 0, -ve integer) of countries a and b, by population.
        public int compare(Country a, Country b) {
            return Long.compare(b.population, a.population);
        }
    }

    // Unit tests the data type.
    public static void main(String[] args) {
        Country[] countries = new Country[5];
        countries[0] = new Country("Brazil", "Brasilia", 218689757);
        countries[1] = new Country("Russia", "Moscow", 141698923);
        countries[2] = new Country("India", "New Delhi", 1425775850);
        countries[3] = new Country("China", "Beijing", 1409670000);
        countries[4] = new Country("South Africa", "Cape Town", 58048332);
        StdOut.println("Unsorted:");
        for (Country country : countries) {
            StdOut.println(country);
        }
        StdOut.println();
        StdOut.println("Sorted by name:");
        Merge.sort(countries);
        for (Country country : countries) {
            StdOut.println(country);
        }
        StdOut.println();
        StdOut.println("Sorted by capital:");
        Merge.sort(countries, Country.capitalOrder());
        for (Country country : countries) {
            StdOut.println(country);
        }
        StdOut.println();
        StdOut.println("Sorted by population:");
        Merge.sort(countries, Country.populationOrder());
        for (Country country : countries) {
            StdOut.println(country);
        }
        StdOut.println();
        StdOut.println("Reverse sorted by population:");
        Merge.sort(countries, Country.reversePopulationOrder());
        for (Country country : countries) {
            StdOut.println(country);
        }
    }
}
