// This library implements merge sort.

import java.util.Comparator;

import stdlib.StdIn;
import stdlib.StdOut;

public class Merge {
    // Sorts the specified array according to the natural ordering of its objects.
    public static <T extends Comparable<T>> void sort(T[] a) {
        T[] aux = a.clone();
        sort(a, aux, 0, a.length - 1);
    }

    // Sorts the specified array according to the order induced by the specified comparator.
    public static <T> void sort(T[] a, Comparator<T> c) {
        T[] aux = a.clone();
        sort(a, aux, 0, a.length - 1, c);
    }

    // Sorts the specified array.
    public static void sort(int[] a) {
        int[] aux = new int[a.length];
        sort(a, aux, 0, a.length - 1);
    }

    // Sorts the specified array.
    public static void sort(double[] a) {
        double[] aux = new double[a.length];
        sort(a, aux, 0, a.length - 1);
    }

    // Sorts a[lo, ..., hi] according to the natural order of its items.
    private static <T extends Comparable<T>> void sort(T[] a, T[] aux, int lo, int hi) {
        if (hi <= lo) {
            return;
        }
        int mid = lo + (hi - lo) / 2;
        sort(a, aux, lo, mid);
        sort(a, aux, mid + 1, hi);
        merge(a, aux, lo, mid, hi);
    }

    // Sorts a[lo, ..., hi] according to the order induced by the comparator c.
    private static <T> void sort(T[] a, T[] aux, int lo, int hi, Comparator<T> c) {
        if (hi <= lo) {
            return;
        }
        int mid = lo + (hi - lo) / 2;
        sort(a, aux, lo, mid, c);
        sort(a, aux, mid + 1, hi, c);
        merge(a, aux, lo, mid, hi, c);
    }

    // Sorts a[lo, ..., hi].
    private static void sort(int[] a, int[] aux, int lo, int hi) {
        if (hi <= lo) {
            return;
        }
        int mid = lo + (hi - lo) / 2;
        sort(a, aux, lo, mid);
        sort(a, aux, mid + 1, hi);
        merge(a, aux, lo, mid, hi);
    }

    // Sorts a[lo, ..., hi].
    private static void sort(double[] a, double[] aux, int lo, int hi) {
        if (hi <= lo) {
            return;
        }
        int mid = lo + (hi - lo) / 2;
        sort(a, aux, lo, mid);
        sort(a, aux, mid + 1, hi);
        merge(a, aux, lo, mid, hi);
    }

    // Merges the two halves a[lo, ..., mid] and a[mid + 1, ..., hi] according to the natural order of its items.
    private static <T extends Comparable<T>> void merge(T[] a, T[] aux, int lo, int mid, int hi) {
        for (int k = lo; k <= hi; k++) {
            aux[k] = a[k];
        }
        int i = lo, j = mid + 1;
        for (int k = lo; k <= hi; k++) {
            if (i > mid) {
                a[k] = aux[j++];
            } else if (j > hi) {
                a[k] = aux[i++];
            } else if (less(aux[j], aux[i])) {
                a[k] = aux[j++];
            } else {
                a[k] = aux[i++];
            }
        }
    }

    // Merges the two halves a[lo, ..., mid] and a[mid + 1, ..., hi] according to order induced by the comparator c.
    private static <T> void merge(T[] a, T[] aux, int lo, int mid, int hi, Comparator<T> c) {
        for (int k = lo; k <= hi; k++) {
            aux[k] = a[k];
        }
        int i = lo, j = mid + 1;
        for (int k = lo; k <= hi; k++) {
            if (i > mid) {
                a[k] = aux[j++];
            } else if (j > hi) {
                a[k] = aux[i++];
            } else if (less(aux[j], aux[i], c)) {
                a[k] = aux[j++];
            } else {
                a[k] = aux[i++];
            }
        }
    }

    // Merges the two halves a[lo, ..., mid] and a[mid + 1, ..., hi].
    private static void merge(int[] a, int[] aux, int lo, int mid, int hi) {
        for (int k = lo; k <= hi; k++) {
            aux[k] = a[k];
        }
        int i = lo, j = mid + 1;
        for (int k = lo; k <= hi; k++) {
            if (i > mid) {
                a[k] = aux[j++];
            } else if (j > hi) {
                a[k] = aux[i++];
            } else if (less(aux[j], aux[i])) {
                a[k] = aux[j++];
            } else {
                a[k] = aux[i++];
            }
        }
    }

    // Merges the two halves a[lo, ..., mid] and a[mid + 1, ..., hi].
    private static void merge(double[] a, double[] aux, int lo, int mid, int hi) {
        for (int k = lo; k <= hi; k++) {
            aux[k] = a[k];
        }
        int i = lo, j = mid + 1;
        for (int k = lo; k <= hi; k++) {
            if (i > mid) {
                a[k] = aux[j++];
            } else if (j > hi) {
                a[k] = aux[i++];
            } else if (less(aux[j], aux[i])) {
                a[k] = aux[j++];
            } else {
                a[k] = aux[i++];
            }
        }
    }

    // Returns true if v is less than w according to their natural order, and false otherwise.
    private static <T extends Comparable<T>> boolean less(T v, T w) {
        return v.compareTo(w) < 0;
    }

    // Returns true if v is less than w according to the order induced by the comparator c, and false otherwise.
    private static <T> boolean less(T v, T w, Comparator<T> c) {
        return c.compare(v, w) < 0;
    }

    // Returns true if v is less than w, and false otherwise.
    private static boolean less(int v, int w) {
        return v < w;
    }

    // Returns true if v is less than w, and false otherwise.
    private static boolean less(double v, double w) {
        return v < w;
    }

    // Unit tests the library.
    public static void main(String[] args) {
        if (args[0].equals("String")) {
            String[] a = StdIn.readAllStrings();
            sort(a);
            for (String s : a) {
                StdOut.print(s + " ");
            }
        } else if (args[0].equals("string")) {
            String[] a = StdIn.readAllStrings();
            sort(a, String.CASE_INSENSITIVE_ORDER);
            for (String s : a) {
                StdOut.print(s + " ");
            }
        } else if (args[0].equals("int")) {
            int[] a = StdIn.readAllInts();
            sort(a);
            for (int i : a) {
                StdOut.print(i + " ");
            }
        } else if (args[0].equals("double")) {
            double[] a = StdIn.readAllDoubles();
            sort(a);
            for (double d : a) {
                StdOut.print(d + " ");
            }
        } else {
            throw new IllegalArgumentException("Illegal command-line argument");
        }
        StdOut.println();
    }
}
