// An iterable data type to represent the First-In-First-Out (FIFO) queue data structure.

import java.util.Iterator;
import java.util.NoSuchElementException;

import stdlib.StdIn;
import stdlib.StdOut;

public class Queue<T> implements Iterable<T> {
    private Node first; // front of the queue
    private Node last;  // back of the queue
    private int n;      // number of items in the queue

    // Initializes an empty queue.
    public Queue() {
        this.first = null;
        this.last = null;
        this.n = 0;
    }

    // Returns true if this queue is empty, and false otherwise.
    public boolean isEmpty() {
        return this.size() == 0;
    }

    // Returns the number of items in this queue.
    public int size() {
        return this.n;
    }

    // Adds item to the end of this queue.
    public void enqueue(T item) {
        Node oldLast = this.last;
        this.last = new Node();
        this.last.item = item;
        if (isEmpty()) {
            this.first = this.last;
        } else {
            oldLast.next = this.last;
        }
        this.n++;
    }

    // Returns the item at the front of this queue.
    public T peek() {
        if (isEmpty()) {
            throw new NoSuchElementException("Queue is empty");
        }
        return this.first.item;
    }

    // Removes and returns the item at the front of this queue.
    public T dequeue() {
        if (isEmpty()) {
            throw new NoSuchElementException("Queue is empty");
        }
        T item = this.first.item;
        this.first = this.first.next;
        this.n--;
        if (isEmpty()) {
            this.last = this.first;
        }
        return item;
    }

    // Returns a string representation of this queue.
    public String toString() {
        String s = "";
        for (T item : this) {
            s += item + ", ";
        }
        return this.isEmpty() ? s + "[]" : "[" + s.substring(0, s.length() - 2) + "]";
    }

    // Returns an iterator that iterates over the items in this queue.
    public Iterator<T> iterator() {
        return new QueueIterator();
    }

    // A data type representing a linked-list of nodes. Each node contains an item and a reference to the next node
    // in the list.
    private class Node {
        private T item;
        private Node next;
    }

    // An iterator over an array.
    private class QueueIterator implements Iterator<T> {
        private Node current; // current item in the iterator

        // Constructs an iterator.
        public QueueIterator() {
            this.current = first;
        }

        // Returns true if there are more items to iterate, and false otherwise.
        public boolean hasNext() {
            return this.current != null;
        }

        // Returns the next item.
        public T next() {
            T item = this.current.item;
            this.current = this.current.next;
            return item;
        }
    }

    // Unit tests the data type.
    public static void main(String[] args) {
        Queue<String> queue = new Queue<String>();
        while (!StdIn.isEmpty()) {
            String item = StdIn.readString();
            if (!item.equals("-")) {
                queue.enqueue(item);
            } else if (!queue.isEmpty()) {
                StdOut.print(queue.dequeue() + " ");
            }
        }
        StdOut.println();
        StdOut.println(queue.size() + " items left in the queue");
        StdOut.println(queue);
    }
}
