// An iterable data type to represent the Last-In-First-Out (LIFO) stack data structure.

import java.util.Iterator;
import java.util.NoSuchElementException;

import stdlib.StdIn;
import stdlib.StdOut;

public class Stack<T> implements Iterable<T> {
    private Node first; // top of the stack
    private int n;      // number of items in the stack

    // Initializes an empty stack.
    public Stack() {
        this.first = null;
        this.n = 0;
    }

    // Returns true if this stack is empty, and false otherwise.
    public boolean isEmpty() {
        return this.size() == 0;
    }

    // Returns the number of items in this stack.
    public int size() {
        return this.n;
    }

    // Adds item to the top of this stack.
    public void push(T item) {
        Node oldFirst = this.first;
        this.first = new Node();
        this.first.item = item;
        this.first.next = oldFirst;
        this.n++;
    }

    // Returns the item at the top of this stack.
    public T peek() {
        if (isEmpty()) {
            throw new NoSuchElementException("Stack is empty");
        }
        return this.first.item;
    }

    // Removes and returns the item at the top of this stack.
    public T pop() {
        if (isEmpty()) {
            throw new NoSuchElementException("Stack is empty");
        }
        T item = this.first.item;
        this.first = this.first.next;
        this.n--;
        return item;
    }

    // Returns a string representation of this stack.
    public String toString() {
        String s = "";
        for (T item : this) {
            s += item + ", ";
        }
        return this.isEmpty() ? s + "[]" : "[" + s.substring(0, s.length() - 2) + "]";
    }

    // Returns an iterator that iterates over the items in this stack.
    public Iterator<T> iterator() {
        return new StackIterator();
    }

    // A data type representing a linked-list of nodes. Each node contains an item and a reference to the next node 
    // in the list.
    private class Node {
        private T item;
        private Node next;
    }

    // Returns an iterator that iterates over the items in this bag.
    private class StackIterator implements Iterator<T> {
        private Node current; // current item in the iterator

        // Constructs an iterator.
        public StackIterator() {
            this.current = first;
        }

        // Returns true if there are more items to iterate, and false otherwise.
        public boolean hasNext() {
            return this.current != null;
        }

        // Returns the next item.
        public T next() {
            T item = this.current.item;
            this.current = this.current.next;
            return item;
        }
    }

    // Unit tests the data type.
    public static void main(String[] args) {
        Stack<String> stack = new Stack<String>();
        while (!StdIn.isEmpty()) {
            String item = StdIn.readString();
            if (!item.equals("-")) {
                stack.push(item);
            } else if (!stack.isEmpty()) {
                StdOut.print(stack.peek() + " ");
                stack.pop();
            }
        }
        StdOut.println();
        StdOut.println(stack.size() + " items left in the stack");
        StdOut.println(stack);
    }
}
