// A data type to represent a symbol table of key-value pairs.

import stdlib.StdOut;

import java.lang.reflect.Array;

public class SymbolTable<K, V> {
    private Node[] st; // an array of chains (linked-lists)
    private int m;     // number of chains in the array
    private int n;     // number of key-value pairs in the symbol table

    // Constructs an empty symbol table.
    @SuppressWarnings("unchecked")
    public SymbolTable() {
        this.m = 4;
        this.st = (Node[]) Array.newInstance(Node.class, this.m);
        this.n = 0;
    }

    // Returns true if this symbol table is empty, and false otherwise.
    public boolean isEmpty() {
        return this.size() == 0;
    }

    // Returns the number of key-value pairs in this symbol table.
    public int size() {
        return this.n;
    }

    // Returns true if this symbol table contains key, and false otherwise.
    public boolean contains(K key) {
        return this.get(key) != null;
    }

    // Returns the value associated with key in this symbol table.
    public V get(K key) {
        if (key == null) {
            throw new IllegalArgumentException("key is null");
        }
        int i = hash(key);
        for (Node x = this.st[i]; x != null; x = x.next) {
            if (key.equals(x.key)) {
                return x.value;
            }
        }
        return null;
    }

    // Inserts a key-value pair into this symbol table.
    public void put(K key, V value) {
        if (key == null) {
            throw new IllegalArgumentException("key is null");
        }
        if (value == null) {
            throw new IllegalArgumentException("value is null");
        }
        if (this.n >= 10 * this.m) {
            resize(2 * this.m);
        }
        int i = hash(key);
        for (Node x = this.st[i]; x != null; x = x.next) {
            if (key.equals(x.key)) {
                x.value = value;
                return;
            }
        }
        Node oldFirst = this.st[i];
        Node first = new Node();
        first.key = key;
        first.value = value;
        first.next = oldFirst;
        this.st[i] = first;
        this.n++;
    }

    // Deletes key and the associated value from this symbol table.
    public void delete(K key) {
        if (key == null) {
            throw new IllegalArgumentException("key is null");
        }
        int i = hash(key);
        this.st[i] = this.delete(this.st[i], key);
        if (this.m > 4 && this.n <= 2 * this.m) {
            resize(this.m / 2);
        }
    }

    // Returns the keys in this symbol table, as an iterable object.
    public Iterable<K> keys() {
        Queue<K> queue = new Queue<K>();
        for (Node chain : this.st) {
            for (Node x = chain; x != null; x = x.next) {
                queue.enqueue(x.key);
            }
        }
        return queue;
    }

    // Returns a string representation of this symbol table.
    public String toString() {
        String s = "";
        for (K key : this.keys()) {
            s += key + ": " + this.get(key) + ", ";
        }
        return this.isEmpty() ? "{}" : "{" + s.substring(0, s.length() - 2) + "}";
    }

    // Deletes key (if it exists) from x, and returns a reference to the (possibly modified) x.
    private Node delete(Node x, K key) {
        if (x == null) {
            return null;
        }
        if (key.equals(x.key)) {
            this.n--;
            return x.next;
        }
        x.next = this.delete(x.next, key);
        return x;
    }

    // A data type representing a linked-list of nodes. Each node contains a key, a value, and a reference to the
    // next node in the list.
    private class Node {
        private K key;
        private V value;
        private Node next;
    }

    // Returns the hash value (an integer between 0 and m - 1) of key.
    private int hash(K key) {
        return (key.hashCode() & 0x7fffffff) % this.m;
    }

    // Resizes the array st to have the given number of chains, rehashing all of the keys.
    @SuppressWarnings("unchecked")
    private void resize(int chains) {
        this.m = chains;
        Node[] temp = (Node[]) Array.newInstance(Node.class, this.m);
        for (Node chain : this.st) {
            for (Node x = chain; x != null; x = x.next) {
                int i = hash(x.key);
                Node oldFirst = temp[i];
                Node first = new Node();
                first.key = x.key;
                first.value = x.value;
                first.next = oldFirst;
                temp[i] = first;
            }
        }
        this.st = temp;
    }

    // Unit tests the data type.
    public static void main(String[] args) {
        String s = "abcdefghijklmnopqrstuvwxyz";
        SymbolTable<Character, Integer> st = new SymbolTable<Character, Integer>();
        StdOut.println("Filling st with characters from the English alphabet ...");
        for (int i = 0; i < s.length(); i++) {
            char c = s.charAt(i);
            st.put(c, i + 1);
        }
        StdOut.println(st);
        StdOut.println("Deleting vowels from st ...");
        String vowels = "aeiou";
        for (int i = 0; i < vowels.length(); i++) {
            st.delete(vowels.charAt(i));
        }
        StdOut.println(st);
    }
}
