// An immutable data type to represent an n-dimensional vector.

import stdlib.StdOut;

public class Vector {
    private int n; // dimension of vector
    private double[] coords; // array of components

    // Constructs a vector given its components.
    public Vector(double[] coords) {
        this.n = coords.length;
        this.coords = new double[n]; // defensive copy of array of components
        for (int i = 0; i < this.n; i++) {
            this.coords[i] = coords[i];
        }
    }

    // Returns the ith component of this vector.
    public double get(int i) {
        return coords[i];
    }

    // Returns the sum of this and the other vector.
    public Vector add(Vector other) {
        double[] result = new double[this.n];
        for (int i = 0; i < this.n; i++) {
            result[i] = this.get(i) + other.get(i);
        }
        return new Vector(result);
    }

    // Returns the difference of this and the other vector.
    public Vector subtract(Vector other) {
        return add(other.scale(-1));
    }

    // Returns the dot product of this and the other vector.
    public double dot(Vector other) {
        double result = 0.0;
        for (int i = 0; i < this.n; i++) {
            result += this.get(i) * other.get(i);
        }
        return result;
    }

    // Returns a scaled (by factor alpha) copy of this vector.
    public Vector scale(double alpha) {
        double[] result = new double[this.n];
        for (int i = 0; i < this.n; i++) {
            result[i] = alpha * this.get(i);
        }
        return new Vector(result);
    }

    // Returns a unit vector in the direction of this vector.
    public Vector direction() {
        return this.scale(1.0 / this.magnitude());
    }

    // Returns the magnitude of this vector.
    public double magnitude() {
        return Math.sqrt(this.dot(this));
    }

    // Returns the dimension of this vector.
    public int dimension() {
        return this.n;
    }

    // Returns a string representation of this vector.
    public String toString() {
        String s = "[";
        for (int i = 0; i < this.n; i++) {
            s += this.get(i) + ", ";
        }
        return this.dimension() == 0 ? s + "]" : s.substring(0, s.length() - 2) + "]";
    }

    // Unit tests the data type.
    public static void main(String[] args) {
        double[] xCoords = {1.0, 2.0, 3.0, 4.0};
        double[] yCoords = {5.0, 2.0, 4.0, 1.0};
        Vector x = new Vector(xCoords);
        Vector y = new Vector(yCoords);
        StdOut.println("x       =  " + x);
        StdOut.println("y       =  " + y);
        StdOut.println("x + y   =  " + x.add(y));
        StdOut.println("x - y   =  " + x.subtract(y));
        StdOut.println("x dot y =  " + x.dot(y));
        StdOut.println("10x     =  " + x.scale(10.0));
        StdOut.println("xhat    =  " + x.direction());
        StdOut.println("|x|     =  " + x.magnitude());
        StdOut.println("ydim    =  " + y.dimension());
    }
}
